
#include <cmath>
#include <mutex>
#include <stdexcept>

std::mutex m;

double f_sqrt(double x){
    if(x<0){
      throw std::invalid_argument("negative numver is not allowed");
    }else{
      return sqrt(x);
    }
}

bool everything_ok(){
    return true;
}

double bad(double x)
{
    m.lock();             // acquire the mutex
    double res = f_sqrt(x);                  // if f() throws an exception, the mutex is never released
    if (!everything_ok())
        return res;           // early return, the mutex is never released
    m.unlock();           // if bad() reaches this statement, the mutex is released
    return -1;
}

double good(double x)
{
    std::lock_guard<std::mutex> lk(m); // RAII class: mutex acquisition is initialization
    double res = f_sqrt(x);            // if f() throws an exception, the mutex is released
    if (!everything_ok()) {
        return res;                    // early return, the mutex is released
    }
    return -1;
}                                      // if good() returns normally, the mutex is released

int main(){
    bad(-2);
    good(-2);
    return 0;
}

